"""OneToOne MapperProperty and Python Property implementations"""

from sqlalchemy import exceptions as sa_exc
from sqlalchemy.sql import and_
from sqlalchemy.orm.interfaces import PropComparator
from sqlalchemy.orm.properties import PropertyLoader, ColumnProperty, ComparableProperty
from sqlalchemy.ext.associationproxy import AssociationProxy

__all__ = [
    'OneToOneProxy', 'OneToOneMapperProperty', 
    'OneToOnePropertyLoader', 'one_to_one'
]

class OneToOneProxy(AssociationProxy):
    """Property that proxies attribute access to columns on a related object.
    
    >>> class Publisher(object):
    ...     def __init__(self, name=None, address=None):
    ...         self.name = name
    ...         self.address = address

    >>> class Book(object):
    ...    def __init__(self, title=None, author=None, publisher=None):
    ...        self.title = title
    ...        self.author = author
    ...        self.publisher = publisher
    ...
    ...    publisher_name = OneToOneProxy('publisher_name', Publisher, 'name')
    ...    publisher_address = OneToOneProxy('publisher_address', Publisher, 'address')

    ... random_house = Publisher(name="Random House", address="NYC")
    ... beloved = Book(title='Beloved', author='Toni Morrison', publisher=random_house)

    ... beloved.publisher is random_house
    True

    .. beloved.publisher.name is beloved.publisher_name
    True
    
    """
    def __init__(self, relation_name, relation_class, attr_name):
        creator = lambda val: relation_class(**{attr_name: val})
        super(OneToOneProxy, self).__init__(relation_name, attr_name, creator=creator)
        # TODO: put this assert in a proper test, 
        # it raises an error if the AssociationProxy 'API' has changed
        assert(self.target_collection == relation_name)
        
    def __get__(self, obj, class_=None):
        """Until Alchemy is patched... """
        if obj is None:
            return self
        target = getattr(obj, self.target_collection, None)
        if target is None:
            return None
        else:
            return super(OneToOneProxy, self).__get__(obj, class_)


class OneToOneMapperProperty(ComparableProperty):
    """Instruments a OneToOneProxy python property for use in query expressions.
    
    Queries on the mapped class will be modified to contain the join condition
    necessary to locate the proxied property.
    
    The property is also registered with the session's unit-of-work machinery,
    so that changes to the property will be persisted.
    
    """
    def __init__(self, proxy_property):
        """Construct a OneToOneMapperProperty.  
        
        proxy_property
            The python descriptor or property to layer comparison behavior on top of.

        """
        # If the python property is not a OneToOneProxy, raise an error
        if not isinstance(proxy_property, OneToOneProxy):
            raise sa_exc.ArgumentError("Unsupported proxy type: '%s'" % type(proxy_property))

        # Use Alchemy's existing ComparableProperty machinery, passing in a callable
        # that will create an appropriate Comparator for this property
        super(OneToOneMapperProperty, self).__init__(one_to_one_comparator, proxy_property)


def one_to_one_comparator(prop, mapper):
    """Return an appropriate OneToOneComparator for the given property."""
    intermediary_obj = mapper.get_property(prop.descriptor.target_collection) # ex: Book.publisher
    target_prop = intermediary_obj.mapper.get_property(prop.key) # ex: Publisher.publisher_name
    
    if isinstance(target_prop, ColumnProperty):
        return OneToOneColumnComparator(prop, mapper, intermediary_obj, target_prop)
    elif isinstance(target_prop, PropertyLoader):
        return OneToOneRelationComparator(prop, mapper, intermediary_obj, target_prop)
    else:
        raise sa_exc.ArgumentError("Unsupported property type: '%s'" % type(prop))


class OneToOneComparator(PropComparator):
    """Base class for other OneToOne Comparators"""
    def __init__(self, prop, mapper, intermediary_obj, target_prop):
        self.intermediary = intermediary_obj # ex: Book.publisher
        self.target = target_prop # ex: Publisher.publisher_name

        # Determine necessary join clauses (ex: book.publisher_id == publisher.publisher_id)
        self.join_clauses = self.intermediary.primaryjoin
        if self.intermediary.secondaryjoin:
            self.join_clauses = and_(self.join_clauses, self.intermediary.secondaryjoin)

        super(OneToOneComparator, self).__init__(prop, mapper)


class OneToOneColumnComparator(OneToOneComparator):
    """Comparator for a OneToOneProxy that's proxying to a ColumnProperty"""
    def __clause_element__(self):
        return self.target.columns[0]._annotate({"parententity": self.intermediary.mapper})

    def operate(self, op, *other, **kwargs):
        """Return 'where' clauses for this operation."""
        join_clauses = self.join_clauses # ex: book.publisher_id = publisher.publisher_id
        op_clauses = op(self.__clause_element__(), *other, **kwargs) # ex: publisher.publisher_name = 'Doubleday'
        return and_(join_clauses, op_clauses) 

    def reverse_operate(self, op, other, **kwargs):
        """Return 'where' clauses for this operation."""
        join_clauses = self.join_clauses # ex: book.publisher_id = publisher.publisher_id
        column = self.__clause_element__() # ex: publisher.publisher_name
        op_clauses = op(column._bind_param(other), column, **kwargs) # ex: 'Doubleday' = publisher.publisher_name
        return and_(join_clauses, op_clauses)


class OneToOneRelationComparator(OneToOneComparator):
    """Comparator for a OneToOneProxy that's proxying to a PropertyLoader property relation"""
    def operate(self, op, *other, **kwargs):
        """Return 'where' clauses for this operation."""
        join_clauses = self.join_clauses # ex: book.publisher_id = publisher.publisher_id
        op_clauses = op(self, *other, **kwargs) # ex: publisher.primary_contact_id == 1
        return sql.and_(join_clauses, op_clauses)

    def reverse_operate(self, op, other, **kwargs):
        return self.operate(self, op, *other, **kwargs)

    def __eq__(self, other):
        # self.target ex: Publisher.primary_contact (relation!)
        if other is None:
            raise sa_exc.InvalidRequestError("NOT IMPLEMENTED!")
        elif self.target.uselist:
            raise sa_exc.InvalidRequestError("NOT IMPLEMENTED!")
        else:
            return self.target._optimized_compare(other)


class OneToOnePropertyLoader(PropertyLoader):
    """PropertyLoader that proxies access to all columns on a related object.
    
    Because this PropertyLoader adds other properties, it must be added
    after the mapper's ``properties`` mapping is already defined, 
    meaning it must be added using ``my_mapper.add_property`` after the
    mapper has already defined.
    """
    def __init__(self, *args, **kwargs):
        kwargs['uselist'] = False
        super(OneToOnePropertyLoader, self).__init__(*args, **kwargs)
    
    def do_init(self):
        super(OneToOnePropertyLoader, self).do_init()
        # Determine the name and class of the proxy object
        relation_name = self.key
        relation_class = self.argument
        # It's likely that no mappers have been compiled yet, so the 
        # properties in self.mapper.iterate_properties won't know their key...
        # SO we have to do something really unpolite to get the list of
        # properties.  Hopefully Alchemy will provide a better way soon...
        for prop_key, prop in self.mapper._props.iteritems():
            # Skip foreign keys (they are already mapped)
            if isinstance(prop, ColumnProperty) and prop.columns[0] in self.remote_side:
                continue

            # Create the Python Property and add it to the mapped class
            proxy_property = OneToOneProxy(relation_name, relation_class, prop_key)

            # Phew! Dynamically adding a descriptor to a class is no walk in the park!
            setattr(self.parent.class_, prop_key, proxy_property)
            
            # Create the MapperProperty so the proxied column will be query-able
            mapper_property = OneToOneMapperProperty(proxy_property)
            
            # Add the MapperProperty to the mapper.  Ideally this would be a
            # call to `self.parent.add_property`, but the mapper is in the
            # middle of compiling, meaning `add_property` doesn't work yet.
            # The `_compile_property` call below forces the added property
            # to initialize.
            self.parent._init_properties[prop_key] = mapper_property
            self.parent._compile_property(prop_key, mapper_property, True)


def one_to_one(*args, **kwargs):
    return OneToOnePropertyLoader(*args, **kwargs)


"""
ReStructured Text -> Google Code Wiki translator
Ian Charnas <ian.charnas@gmail.com>
Modified from the rst2wiki.py from the python-nose project
"""

import sys
from docutils.nodes import SparseNodeVisitor, paragraph, title_reference, \
    emphasis
from docutils.writers import Writer
from docutils.core import publish_string

class WikiWriter(Writer):
    def translate(self):
        visitor = WikiVisitor(self.document)
        self.document.walkabout(visitor)
        self.output = visitor.astext()

class WikiVisitor(SparseNodeVisitor):

    def __init__(self, document):
        SparseNodeVisitor.__init__(self, document)
        self.list_depth = 0
        self.list_item_prefix = None
        self.indent = self.old_indent = ''
        self.output = []
        self.preformat = False
        
    def astext(self):
        # ICC Using google code wiki, don't need >>> and <<<
        return ''.join(self.output)
        # return '\n>>>\n\n'+ ''.join(self.output) + '\n\n<<<\n'

    def visit_Text(self, node):
        #print "Text", node
        data = node.astext()
        if not self.preformat:
            data = data.lstrip('\n\r')
            data = data.replace('\r', '')
            data = data.replace('\n', ' ')
        self.output.append(data)
    
    def visit_bullet_list(self, node):
        self.list_depth += 1
        self.list_item_prefix = (' ' * self.list_depth) + '* '

    def depart_bullet_list(self, node):
        self.list_depth -= 1
        if self.list_depth == 0:
            self.list_item_prefix = None
        else:
            (' ' * self.list_depth) + '* '
        self.output.append('\n\n')
                           
    def visit_list_item(self, node):
        self.old_indent = self.indent
        self.indent = self.list_item_prefix

    def depart_list_item(self, node):
        self.indent = self.old_indent
        
    def visit_literal_block(self, node):
        self.output.extend(['{{{', '\n'])
        self.preformat = True

    def depart_literal_block(self, node):
        self.output.extend(['\n', '}}}', '\n\n'])
        self.preformat = False
        
    def visit_paragraph(self, node):
        self.output.append(self.indent)
        
    def depart_paragraph(self, node):
        self.output.append('\n\n')
        if self.indent == self.list_item_prefix:
            # we're in a sub paragraph of a list item
            self.indent = ' ' * self.list_depth
        
    def visit_reference(self, node):
        # for key in dir(node):
        #     val = getattr(node, key, None)
        #     print "node.%s = %r" % (key, val)
        # import sys
        # sys.exit(-1)
        # 
        if node.has_key('refuri'):
            href = node['refuri']
        elif node.has_key('refid') and node.has_key('name'):
            href = '#' + node['name'].replace(" ", "_")
            #href = '#' + node['refid']
        else:
            href = None
        self.output.append('[' + href + ' ')

    def depart_reference(self, node):
        self.output.append(']')

    def visit_subtitle(self, node):
        self.output.append('=== ')

    def depart_subtitle(self, node):
        self.output.append(' ===\n\n')
        self.list_depth = 0
        self.indent = ''
        
    def visit_title(self, node):
        self.output.append('== ')

    def depart_title(self, node):
        self.output.append(' ==\n\n')
        self.list_depth = 0
        self.indent = ''
        
    def visit_title_reference(self, node):
        self.output.append("`")

    def depart_title_reference(self, node):
        self.output.append("`")

    def visit_emphasis(self, node):
        self.output.append('*')

    def depart_emphasis(self, node):
        self.output.append('*')
        
    def visit_literal(self, node):
        self.output.append('`')
        
    def depart_literal(self, node):
        self.output.append('`')
   
        
def main(source):
    output = publish_string(source, writer=WikiWriter())
    print output
    
if __name__ == '__main__':
    main(sys.stdin.read())



"""Release information and release tools"""
import os
from textwrap import dedent
from distutils import log
from distutils.core import Command
from distutils.errors import DistutilsOptionError

# Define pypi release information
NAME = "BrainFreeze"
VERSION = "0.1rc2"
DESCRIPTION = dedent("""
    BrainFreeze is an SQLAlchemy plugin that supports object composition (as opposed to 
    class inheritance).  Google for "composition vs inheritance" to get the general idea.
    
    Another way to think of BrainFreeze is that it's a plugin for proxying properties on
    one-to-one related objects.  BrainFreeze uses SQLAlchemy's AssociationProxy for the
    python properties, only BrainFreeze goes the extra step of making those properties
    query-able using SQLAlchemy's session.query() machinery.
""")
AUTHOR = "Ian Charnas, Brian Beck"
AUTHOR_EMAIL = "ian.charnas@gmail.com, exogen@gmail.com"
URL = "http://brainfreeze-alchemy.googlecode.com/"
DOWNLOAD_URL = "http://code.google.com/p/brainfreeze-alchemy/downloads/list"

# Add support for `python setup.py googlecode_upload`
#
# Note: This is not a very generalized way to do this, clearly I should be 
# using user_options instead of relying on some constants
GOOGLECODE_PROJECT = "brainfreeze-alchemy"
from brainfreeze.release.googlecode_upload import upload_find_auth
class googlecode_upload(Command):
    """Setuptools Command to upload packages to google code"""
    description = "upload packages to google code"
    
    user_options = []
    
    def initialize_options(self):
        pass
        
    def finalize_options(self):
        pass

    def run(self):
        if not self.distribution.dist_files:
            raise DistutilsOptionError(
                "No dist file created in earlier command"
            )
        for command, pyversion, filename in self.distribution.dist_files:
            self.upload_file(command, pyversion, filename)
            
    def upload_file(self, command, pyversion, filename):
        command_filetype_map = {
            'bdist_egg': ("Python Egg", ['Type-Package', 'OpSys-All']),
            'bdist_msi': (
                "Microsoft Installer Package", 
                ['Type-Package', 'OpSys-Windows']
            ),
            'bdist_rpm': ("RPM", ['Type-Package', 'OpSys-Linux']),
            'bdist_wininst': (
                "Auto-Installing Executable for Microsoft Windows", 
                ['Type-Installer', 'OpSys-Windows']
            ),
            'sdist': ("Source Distribution", ['Type-Source', 'OpSys-All'])
        }
        
        self.announce("Submitting %s to Google Code" % filename, log.INFO)
        try:
            status, reason, url = upload_find_auth(
                filename, # file to upload
                GOOGLECODE_PROJECT, # which project to upload file to?
                command_filetype_map[command][0], # summary of file
                command_filetype_map[command][1] # labels for file
            )
            if url:
                self.announce(
                    "Google Code upload successful to URL: %s" % url
                )
            else:
                self.announce(
                    "Google Code upload failed.  Server said %s (%s)" %
                    (reason, status)
                )
        except Exception, e:
            self.announce("Upload failed: %s" % str(e), log.ERROR)      
